import collections
import hashlib
import copy

class WeChatSignHelper:
    def __init__(self,dataDict,apiKey):
        self.data = copy.deepcopy(dataDict) #type: dict
        self.apiKey = apiKey
        self.keyValueString = ""
        self.clean_up()
        self.dataToKeyValueString()


    def clean_up(self):
        if "sign" in self.data:
            del self.data["sign"]

    def dataToKeyValueString(self):
        od = collections.OrderedDict(sorted(self.data.items()))
        for key,value in od.items():
            self.keyValueString+=f"{key}={value}&"
        self.keyValueString += f"key={self.apiKey}"

    def getSign(self):
        md5Obj = hashlib.md5()
        md5Obj.update(self.keyValueString.encode("utf-8"))
        return md5Obj.hexdigest().upper()









import base64
import json
from Crypto.Cipher import AES

class WXBizDataCrypt:
    def __init__(self, appId, sessionKey):
        self.appId = appId
        self.sessionKey = sessionKey

    def decrypt(self, encryptedData, iv):
        # base64 decode
        sessionKey = base64.b64decode(self.sessionKey)
        encryptedData = base64.b64decode(encryptedData)
        iv = base64.b64decode(iv)

        cipher = AES.new(sessionKey, AES.MODE_CBC, iv)
        result=cipher.decrypt(encryptedData)
        unpaded=self._unpad(result)
        decrypted = json.loads(unpaded)

        if decrypted['watermark']['appid'] != self.appId:
            raise Exception('Invalid Buffer')

        return decrypted

    def _unpad(self, s):
        return s[:-ord(s[len(s)-1:])]



from django.contrib.auth.models import User
from django.db import models

# Create your models here.
from django.db.models.signals import post_save
from django.dispatch import receiver


class WeappUser(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE,related_name='weapp_user')
    nickName = models.CharField(max_length=50, blank=True, default='NO NICKNAME')
    avatarUrl = models.TextField(blank=True, null=True)
    gender = models.IntegerField(blank=True, null=True)
    city = models.CharField(max_length=100, blank=True, default='NO CITY')
    province = models.CharField(max_length=100, blank=True, default='NO PROVINCE')
    country = models.CharField(max_length=100, blank=True, default='NO COUNTRY')
    language = models.CharField(max_length=20, blank=True, default='NO LANG')
    unionId = models.CharField(max_length=50, blank=True, null=True)
    sessionKey = models.CharField(max_length=50, blank=True, null=True)
    token = models.CharField(max_length=32, null=True)


@receiver(post_save, sender=User)
def create_weapp_user(sender, instance, created, **kwargs):
    if created:
        WeappUser.objects.create(user=instance)


@receiver(post_save, sender=User)
def update_weapp_user(sender, instance, **kwargs):
    instance.weapp_user.save()



# Generated by Django 2.2.1 on 2019-07-14 07:02

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='WeappUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nickName', models.CharField(blank=True, default='NO NICKNAME', max_length=50)),
                ('avatarUrl', models.TextField(blank=True, null=True)),
                ('gender', models.IntegerField(blank=True, null=True)),
                ('city', models.CharField(blank=True, default='NO CITY', max_length=100)),
                ('province', models.CharField(blank=True, default='NO PROVINCE', max_length=100)),
                ('country', models.CharField(blank=True, default='NO COUNTRY', max_length=100)),
                ('language', models.CharField(blank=True, default='NO LANG', max_length=20)),
                ('unionId', models.CharField(blank=True, max_length=50, null=True)),
                ('sessionKey', models.CharField(blank=True, max_length=50, null=True)),
                ('token', models.CharField(max_length=32, null=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='weapp_user', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from django.utils.timezone import now

from rest_framework import authentication, exceptions
import logging

from weapp_auth.models import WeappUser

logger=logging.getLogger(__name__)
SAFE_METHODS = ('GET', 'HEAD', 'OPTIONS')

class OpenIdAuthenticator(authentication.BaseAuthentication):
    def authenticate(self, request):
        token = request.META.get('HTTP_TOKEN')
        logger.info("OpenId Authenticator token = "+str(token))
        if not token:
            return None

        weapp_user = WeappUser.objects.filter(token=token).first()
        if not weapp_user:
            raise exceptions.AuthenticationFailed('Authentication failed, no such user')
        user=weapp_user.user
        # if user.is_staff:
        #     raise exceptions.AuthenticationFailed('Authentication failed, not admin authenticator')

        user.last_login=now()
        logger.info("user {} last login {} updated".format(user.username,user.last_login))


        return (user, token)



import logging

import requests
from django.shortcuts import render

# Create your views here.
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from django.conf import settings

from weapp_auth.authenticator.OpenIdAuthenticator import OpenIdAuthenticator
from weapp_auth.wechat.WeChatUserHelper import handleWechatLogin, handleUpdateUserInfo

logging.basicConfig()
logger = logging.getLogger("weapp_auth_views")
logger.setLevel(logging.DEBUG)

WECHAT_CODE_TO_SESSION_URL="https://api.weixin.qq.com/sns/jscode2session?appid={}&secret={}&js_code={}&grant_type=authorization_code"

class WeChatLoginApiView(APIView):
    authentication_classes = [OpenIdAuthenticator]
    permission_classes = []
    def post(self,request,format=None):
        code=request.data["code"]
        logger.debug(f"wechat login code {code}")
        if not code:
            return Response(status.HTTP_400_BAD_REQUEST)
        url=WECHAT_CODE_TO_SESSION_URL.format(settings.WEAPP_AUTH['APP_ID'],settings.WEAPP_AUTH['APP_SECRET'],code)
        resp=requests.get(url)
        respJson=resp.json()
        unionId=None
        logger.debug(f"resp from wechat server {resp}")
        logger.debug(f"resp json from wechat server {respJson}")
        if "unionid" in respJson:
            unionId=respJson["unionid"]
        token,userId=handleWechatLogin(respJson["openid"],respJson["session_key"],unionId)
        return Response(data={"token":token,"id":userId},status=status.HTTP_200_OK)

class WeChatUpdateUserInfoApiView(APIView):
    authentication_classes = [OpenIdAuthenticator]
    permission_classes = [IsAuthenticated]
    def post(self,request,format=None):
        try:
            handleUpdateUserInfo(request.data["encryptedData"],request.data["iv"],request.user)
        except Exception as ex:
            return Response(data={"error":str(ex)},status=status.HTTP_400_BAD_REQUEST)
        return Response(data={"id":self.request.user.id}, status=status.HTTP_200_OK)



from django.contrib.auth.models import User
from django.conf import settings
from weapp_auth.wechat.WXBizDataCrypt import WXBizDataCrypt
import uuid
import logging

logger=logging.getLogger(__name__)


def handleWechatLogin(openid:str,session_key:str,unionid:str):
    user=initOrGetUser(openid)
    user.weapp_user.sessionKey=session_key
    user.weapp_user.unionId=unionid
    token=str(uuid.uuid4()).replace('-','')
    user.weapp_user.token=token
    user.save()
    return token,user.id

def handleUpdateUserInfo(encryptedData:str,iv:str,user:User):
    wxCrypt=WXBizDataCrypt(settings.WEAPP_AUTH['APP_ID'],user.weapp_user.sessionKey)
    data=wxCrypt.decrypt(encryptedData,iv)
    if user.username!=data["openId"]:
        logger.error("User mismatch while update user info incoming openid={} db openid={}".format(data["openId"],user.username))
        raise Exception("User mismatch")
    user.weapp_user.nickName=data["nickName"]
    user.weapp_user.gender=data["gender"]
    user.weapp_user.city=data["city"]
    user.weapp_user.province=data["province"]
    user.weapp_user.country=data["country"]
    user.weapp_user.avatarUrl=data["avatarUrl"]
    if "unionId" in data:
        user.weapp_user.unionId=data["unionId"]
    user.weapp_user.language=data["language"]
    user.save()


def initOrGetUser(openid):
    user,created=User.objects.get_or_create(username=openid)
    return user


from django.conf.urls import url
from django.urls import include, path
from rest_framework import routers

from mi_django_rest_starter import settings
from weapp_auth import views

urlpatterns = [
    url(r'^wechatLogin/', views.WeChatLoginApiView.as_view()),
    url(r'^wechatUpdateUserInfo/', views.WeChatUpdateUserInfoApiView.as_view()),
]

